# Partsd of the code are adopted from 
# M. Wolters. A genetic algorithm for selection of fixed-size subsets with application to design
# problems. Journal of Statistical Software, Code Snippets, 68(1):1–18, 2015.

# Design matrix for the 4-factor problem 
AllRuns4 <- matrix(nrow = 81, ncol = 4)
AllRuns4[, 1] <- rep(c(-1, 0, 1), each = 3^3)
AllRuns4[, 2] <- rep(rep(c(-1, 0, 1), each = 3^2), 3)
AllRuns4[, 3] <- rep(rep(c(-1, 0, 1), each = 3), 9)
AllRuns4[, 4] <- rep(c(1, 2, 3), 3^3)

MakeModel4 <- function(AllRuns) {
  X <- matrix(nrow = dim(AllRuns)[1], ncol = 18)
  X[, 1] <- 1                                   
  X[, 2:4] <- AllRuns[, 1:3]                    
  
  X[AllRuns[, 4] == 1, 5] <- sqrt(6)/2
  X[AllRuns[, 4] == 1, 6] <- -1/sqrt(2)
  X[AllRuns[, 4] == 2, 5] <- 0
  X[AllRuns[, 4] == 2, 6] <- sqrt(2)
  X[AllRuns[, 4] == 3, 5] <- -sqrt(6)/2
  X[AllRuns[, 4] == 3, 6] <- -1/sqrt(2)

  X[, 7] <- AllRuns[, 1] * AllRuns[, 2]         
  X[, 8] <- AllRuns[, 1] * AllRuns[, 3]         
  X[, 9] <- AllRuns[, 2] * AllRuns[, 3]         
  X[, 10:12] <- AllRuns[, 1:3]^2                
  X[, 13] <- AllRuns[, 1] * X[, 5]              
  X[, 14] <- AllRuns[, 1] * X[, 6]              
  X[, 15] <- AllRuns[, 2] * X[, 5]              
  X[, 16] <- AllRuns[, 2] * X[, 6]              
  X[, 17] <- AllRuns[, 3] * X[, 5]              
  X[, 18] <- AllRuns[, 3] * X[, 6]              
  X
}



n <<- 81
k <<- 24
X <- MakeModel4(AllRuns4)
M <<- X

library("kofnGA")
library("DEoptim")
library("pso")
library("GA")
library("CEoptim")
library(tictoc)

ObjFunMax <- function(x){
  ids <- order(x)
  idx <- ids[1:k]
  A <- M[idx, ]
  B <- t(A) %*% A
  B <- (B + t(B))/2  # Ccorrect round-off errors
  return((determinant(B, log = TRUE)$modulus))} 

ObjFunMin <- function(x){
  ids <- order(x)
  idx <- ids[1:k]
  A <- M[idx, ]
  B <- t(A) %*% A
  B <- (B + t(B))/2  # Correct round-off errors
  return(-(determinant(B, log = TRUE)$modulus))} 

ObjFunMinCE <- function(x){
  ids <- order(x[1:81])
  idx <- ids[1:k]
  A <- M[idx, ]
  B <- t(A) %*% A
  B <- (B + t(B))/2  #to ensure symmetric (correct round-off errors)
  return(-(determinant(B, log = TRUE)$modulus))} 


ubnd = rep.int(1, n)
lbnd = rep.int(0, n)

repNum = 20

#############################################################
# kofnGA
#############################################################


# The objective function (D-optimality criterion, for use in the GA search)
ObjFun <- function(idx, Matrix) {
  A <- Matrix[idx, ]
  B <- t(A) %*% A
  B <- (B + t(B))/2  #to ensure symmetric (correct round-off errors)
  -(determinant(B, log = TRUE)$modulus)
}

set.seed(12345)
for(i in 1:repNum) {
  ptm <- proc.time()

  result <- kofnGA(n = 81, k = 24,  
                   OF = ObjFun, keepbest = 50, ngen=4750,
                   tourneysize = 50, Matrix = M)
  t = proc.time() - ptm
  
  print(paste0(result$bestobj," ",t[3]))
}  
  

#############################################################
# Differential Evolution
#############################################################
set.seed(12345)
print("DE")
print("----------------------------------------------------")
for(i in 1:repNum) {
  ptm <- proc.time()
  outDEoptim <- DEoptim(ObjFunMin, lbnd, ubnd,                         
                        DEoptim.control(NP = 10*n, itermax = 2000,
                                        trace=FALSE,strategy=2))
  t = proc.time() - ptm
  print(paste0(outDEoptim$optim$bestval," ",t[3]))
}

#############################################################
# PSO
#############################################################
set.seed(12345)
print("PSO")
print("----------------------------------------------------")
for(i in 1:repNum) {
  ptm <- proc.time()
  outPSOoptim <- psoptim(rep(NA,n), ObjFunMin, 
                         lower = lbnd, upper = ubnd,
                         control=list(trace=0,s = 10*n, maxit=850))
  t = proc.time() - ptm
  print(paste0(outPSOoptim$value," ",t[3]))
}
print("----------------------------------------------------")

#############################################################
# GA
#############################################################
set.seed(12345)
print("GA")
print("----------------------------------------------------")
for(i in 1:repNum) {
  ptm <- proc.time()

GA <- ga(type = "real-valued", fitness = ObjFunMax, 
         lower = lbnd, popSize = 6*n, maxiter = 800, monitor=1, upper = ubnd)
         
t = proc.time() - ptm
print(paste0(attr(GA, "fitnessValue")," ",t[3] ))
}

#############################################################
# CE
#############################################################

set.seed(12345)
mu0 <- rep(0,n); sigma0 <- rep(1,n)
print("CE")
print("----------------------------------------------------")
for(i in 1:repNum) {
  ptm <- proc.time()
res <- CEoptim(ObjFunMinCE,  
               continuous=list(mean=mu0, sd=sigma0,
                               smoothMean=0.7, smoothSd=0.7),
               maximize=FALSE, N=30*n*k, rho=0.1,
               iterThr=1e4L,noImproveThr=5, verbose=FALSE)
t = proc.time() - ptm
print(paste0(res$optimum," ",t[3] ))
}
